# Copyright 2020 PureHing
# Copyright 2021 NXP
#
# Parts of this file were copied from PureHing's github repository
# respecting its rights. All the modified parts below are according to
# MIT LICENSE terms.
#
# SPDX-License-Identifier:    MIT
# Reference:
# https://github.com/PureHing/face-mask-detection-tf2/blob/master/LICENSE

import os
import cv2
import numpy as np
from PIL import Image

from eiq.config import FONT
from eiq.engines.tflite.inference import TFLiteInterpreter
from eiq.modules.detection.config import LEFT, TOP, RIGHT, BOTTOM, CONFIDENCE, CLASSES, OBJ_COVID19
from eiq.modules.utils import DemoBase


class eIQCOVID19Detection(DemoBase):
    def __init__(self, args=None):
        super().__init__(args, self.__class__.__name__, OBJ_COVID19)

        self.social_distance_model = None
        self.facemask_model = None
        self.interpreter_social_distance = None
        self.interpreter_facemask = None

        self.mode = None

        #camera info
        self.eye_height = None
        self.eye_angle = None
        self.eye_slope = None

        self.label_person =  ['???','person']
        #facemask
        self.labels_facemask =  {0: 'Mask', 1: 'NoMask'}
        self.priors = None
        self.count = 0

        self.run_ssd = 1
        self.run_facemask = 0
        self.result_ssd = None
        self.result_face = None

    @staticmethod
    def description():
        return ("This demo uses:\n   * TensorFlow Lite as inference engine."
                "\n   * Single Shot Detection as default algorithm.\n")

    def prepare_params(self):
        flag = None

        if hasattr(self.args, 'work_mode'):
            flag = self.args.work_mode

        priors_file = os.path.join(self.model_dir, 'priors.txt')
        self.priors = np.reshape(np.loadtxt(priors_file), (9949, 4))

        #set self.model
        self.social_distance_model = os.path.join(self.model_dir, "mobilenet_ssd_v2_coco_quant_postprocess.tflite")
        self.facemask_model = os.path.join(self.model_dir, "facemask_int8.tflite")
        #set cam param
        camera_params = []
        if hasattr(self.args, 'camera_params'):
            camera_params = self.args.camera_params.split()
            self.eye_height = int(camera_params[0])
            self.eye_angle = int(camera_params[1])
            self.eye_slope = int(camera_params[2])
        else:
            self.eye_height = 200
            self.eye_angle = 60
            self.eye_slope = 60

        """
        SSD FACEMASK IMAGE VIDEO CAMERA

        """
        if flag == 'None':
            self.mode = 0
        elif flag == '1':
            self.mode = 1
        else :
            self.mode = 0

    def usage(self, name=None, labels=True, model=True):
        super().usage(name=name, labels=labels, model=model)


    def softmax(self, x):
        orig_shape = x.shape
        #print("orig_shape", orig_shape)

        if len(x.shape) > 1:
            # matrix
            tmp = np.max(x, axis=1)
            x -= tmp.reshape((x.shape[0], 1))
            x = np.exp(x)
            tmp = np.sum(x, axis=1)
            x /= tmp.reshape((x.shape[0], 1))
            #print("matrix")
        else:
            # vector
            tmp = np.max(x)
            x -= tmp
            x = np.exp(x)
            tmp = np.sum(x)
            x /= tmp
            #print("vector")
        return x

    def social_distant_infer(self, image):
        self.interpreter_social_distance.set_tensor(np.expand_dims(image, axis=0))
        self.interpreter_social_distance.run_inference()

        positions = self.interpreter_social_distance.get_tensor(0, squeeze=True)
        classes = self.interpreter_social_distance.get_tensor(1, squeeze=True)
        scores = self.interpreter_social_distance.get_tensor(2, squeeze=True)

        classes = np.squeeze(classes+1).astype(np.int32)

        result = []
        for idx, score in enumerate(scores):
            if (score > 0.5) and (classes[idx] == 1):
                result.append({'pos': positions[idx], '_id': classes[idx]})
        return result

    def draw_result(self, frame, result):
        width = frame.shape[1]
        height = frame.shape[0]

        GPS = []

        for obj in result:
            pos = obj['pos']
            _id = obj['_id']

            x1 = int(pos[1] * width)
            x2 = int(pos[3] * width)
            y1 = int(pos[0] * height)
            y2 = int(pos[2] * height)

            top = max(0, np.floor(y1 + 0.5).astype('int32'))
            left = max(0, np.floor(x1 + 0.5).astype('int32'))
            bottom = min(height, np.floor(y2 + 0.5).astype('int32'))
            right = min(width, np.floor(x2 + 0.5).astype('int32'))

            #method 1, not calculate > 90° range
            angle_y =(1 - bottom / height) * self.eye_angle + (self.eye_slope - self.eye_angle / 2)
            if (angle_y >= 90):
                continue

            angle_x = ((left + right) / 2 / width - 0.5) * (self.eye_angle / 2)
            Y = self.eye_height * np.tan(angle_y * np.pi / 180)
            X = Y * np.tan(angle_x * np.pi / 180) * 1.4
            GPS.append({'top':top, 'left': left, 'bottom':bottom,'right':right, 'X':X, 'Y':Y, 'warning': False})
            # end method 1

        return GPS

    def draw_result_ssd(self, frame, GPS):

        if GPS == None:
            return
        #calculate social distance
        persons = len(GPS)
        for index_1 in range(persons-1):
            for j  in range(persons - index_1 - 1):
                index_2 = index_1 + j + 1
                if abs(GPS[index_2]['X'] - GPS[index_1]['X']) < 100 and abs(GPS[index_2]['Y'] - GPS[index_1]['Y']) < 100:
                    rect1_x = (int)((GPS[index_1]['right'] + GPS[index_1]['left']) / 2)
                    rect1_y = (int)((GPS[index_1]['top'] + GPS[index_1]['bottom']) / 2)
                    rect2_x = (int)((GPS[index_2]['right'] + GPS[index_2]['left']) / 2)
                    rect2_y = (int)((GPS[index_2]['top'] + GPS[index_2]['bottom']) / 2)

                    GPS[index_1]['warning'] = True
                    GPS[index_2]['warning'] = True

                    cv2.line(frame,(rect1_x,rect1_y),(rect2_x,rect2_y),(0,0,255), 3)

        for index in range(persons):
            if GPS[index]['warning'] == True:
                color = FONT['color']['red']
            else:
                color = FONT['color']['green']

            left = (int)(GPS[index]['left'])
            top  = (int)(GPS[index]['top'])
            right = (int)(GPS[index]['right'])
            bottom  = (int)(GPS[index]['bottom'])

            cv2.rectangle(frame, (left, top), (right, bottom),
                          color, 6)


        self.overlay.draw_info(frame, self.social_distance_model, self.media_src,
                                   self.interpreter_social_distance.inference_time,
                                   self.interpreter_social_distance.input_details[0]['quantization'])

    def decode_bbox(self, bbox, priors, variances):
        """
        Decode locations from predictions using anchors to undo
        the encoding we did for offset regression at train time.
        """
        if variances is None:
            variances = [0.1, 0.2]
        boxes = np.concatenate(
            (priors[:, :2] + bbox[:, :2] * variances[0] * priors[:, 2:],
            priors[:, 2:] * np.exp(bbox[:, 2:] * variances[1])), 1)
        boxes[:, :2] -= boxes[:, 2:] / 2
        boxes[:, 2:] += boxes[:, :2]
        return boxes

    def single_class_non_max_suppression(self, bboxes, confidences, conf_thresh=0.6, iou_thresh=0.5, keep_top_k=-1):
        '''
        do nms on single class.
        Hint: for the specific class, given the bbox and its confidence,
        1) sort the bbox according to the confidence from top to down, we call this a set
        2) select the bbox with the highest confidence, remove it from set, and do IOU calculate with the rest bbox
        3) remove the bbox whose IOU is higher than the iou_thresh from the set,
        4) loop step 2 and 3, util the set is empty.
        :param bboxes: numpy array of 2D, [num_bboxes, 4]
        :param confidences: numpy array of 1D. [num_bboxes]
        :param conf_thresh:
        :param iou_thresh:
        :param keep_top_k:
        :return:
        '''
        if len(bboxes) == 0: return []

        conf_keep_idx = np.where(confidences > conf_thresh)[0]

        bboxes = bboxes[conf_keep_idx]
        confidences = confidences[conf_keep_idx]

        pick = []
        xmin = bboxes[:, 0]
        ymin = bboxes[:, 1]
        xmax = bboxes[:, 2]
        ymax = bboxes[:, 3]

        area = (xmax - xmin + 1e-3) * (ymax - ymin + 1e-3)
        idxs = np.argsort(confidences)

        while len(idxs) > 0:
            last = len(idxs) - 1
            i = idxs[last]
            pick.append(i)

            # keep top k
            if keep_top_k != -1:
                if len(pick) >= keep_top_k:
                    break

            overlap_xmin = np.maximum(xmin[i], xmin[idxs[:last]])
            overlap_ymin = np.maximum(ymin[i], ymin[idxs[:last]])
            overlap_xmax = np.minimum(xmax[i], xmax[idxs[:last]])
            overlap_ymax = np.minimum(ymax[i], ymax[idxs[:last]])
            overlap_w = np.maximum(0, overlap_xmax - overlap_xmin)
            overlap_h = np.maximum(0, overlap_ymax - overlap_ymin)
            overlap_area = overlap_w * overlap_h
            overlap_ratio = overlap_area / (area[idxs[:last]] + area[i] - overlap_area)

            need_to_be_deleted_idx = np.concatenate(([last], np.where(overlap_ratio > iou_thresh)[0]))
            idxs = np.delete(idxs, need_to_be_deleted_idx)

        # if the number of final bboxes is less than keep_top_k, we need to pad it.
        # TODO
        return conf_keep_idx[pick]


    def facemask_infer(self, input_data, image_pad, image_raw):

        height_pad, width_pad, _ = image_pad.shape
        height, width, _ = image_raw.shape
        variances = [0.1, 0.2]
        result = []

        self.interpreter_facemask.set_tensor(np.expand_dims(input_data, axis=0))
        self.interpreter_facemask.run_inference()

        predictions = self.interpreter_facemask.get_tensor(0, squeeze=False)


        bbox_regressions, confs = np.split(predictions[0], [4, ], axis=-1)
        boxes = self.decode_bbox(bbox_regressions, self.priors, variances)

        confs = self.softmax(confs)
        confs = confs[:, [1, 2]]

        bbox_max_scores = np.max(confs, axis=1)
        bbox_max_score_classes = np.argmax(confs, axis=1)

        keep_idxs = self.single_class_non_max_suppression(boxes,
                                                 bbox_max_scores,
                                                 conf_thresh=0.5,
                                                 iou_thresh=0.5,
                                                 )

        ######### draw image
        for idx in keep_idxs:
            conf = float(bbox_max_scores[idx])
            class_id = bbox_max_score_classes[idx]
            bbox = boxes[idx]
            # clip the coordinate, avoid the value exceed the image boundary.
            xmin = max(0, int(bbox[0] * width_pad))
            ymin = max(0, int(bbox[1] * height_pad))
            xmax = min(int(bbox[2] * width_pad), width)
            ymax = min(int(bbox[3] * height_pad), height)

            result.append([xmin,ymin,xmax,ymax,class_id])

        return result

    def draw_result_face(self, frame, result):

        if result == None:
            return

        for idx in result:
            if idx[4] == 0:
                color = (0, 255, 0)
            else:
                color = (0, 0, 255)
            cv2.rectangle(frame, (idx[0], idx[1]), (idx[2], idx[3]), color, 2)
            cv2.putText(frame, "%s" % (self.labels_facemask[idx[4]]), (idx[0] + 2, idx[1] - 2),
                        cv2.FONT_HERSHEY_SIMPLEX, 0.8, color)

        self.overlay.draw_info(frame, self.facemask_model, self.media_src,
                                   self.interpreter_facemask.inference_time,
                                   self.interpreter_facemask.input_details[0]['quantization'])

    def pad_input_image(self, img):
        """pad image to suitable shape"""
        img_h, img_w, _ = img.shape

        img_pad_h = 0
        img_pad_w = 0

        if img_w > img_h:
            img_pad_h = img_w - img_h
        else:
            img_pad_w = img_h - img_w

        img = cv2.copyMakeBorder(img, 0, img_pad_h, 0, img_pad_w,
                             cv2.BORDER_CONSTANT)
        return img

    def image_pre_process(self, image_raw):

        img_f = np.float32(image_raw.copy())
        img_pad = self.pad_input_image(img_f)
        img_resize = cv2.resize(img_pad,(self.interpreter_facemask.width(), self.interpreter_facemask.height()))
        img_data =  img_resize / 255.0 - 0.5
        img = cv2.cvtColor(img_data, cv2.COLOR_BGR2RGB)

        return img, img_pad

    def detect_object(self, frame):


        if self.mode == 1:
            self.count += 1
            self.run_ssd = self.count % 2
            self.run_facemask = (self.count + 1) % 2
        else :
            self.run_ssd = 0
            self.run_facemask = 1

        if self.run_ssd:
            image = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            image = image.resize((self.interpreter_social_distance.width(), self.interpreter_social_distance.height()))
            result = self.social_distant_infer(image)
            self.result_ssd = self.draw_result(frame, result)

        if self.run_facemask:
            facemask_input, facemask_pad = self.image_pre_process(frame)
            self.result_face = self.facemask_infer(facemask_input, facemask_pad, frame)

        self.draw_result_ssd(frame, self.result_ssd)
        self.draw_result_face(frame, self.result_face)


        if self.count == 1000:
            self.count = 0

        return frame

    def start(self):
        self.gather_data()
        self.prepare_params()
        self.interpreter_social_distance = TFLiteInterpreter(self.social_distance_model)
        self.interpreter_facemask = TFLiteInterpreter(self.facemask_model)


    def run(self):
        self.start()
        self.run_inference(self.detect_object)



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os


VIDEO_SWITCH_CORE = {'sha1': "aa8864498fed9294de27b273a2cce6a6a9097726",
                     'src': {'drive': "https://drive.google.com/file/d/"
                                      "1p3hVhEt_nFuiBw5Sj2EI7ghxmYWdggnh",
                             'github': "https://github.com/NXP/"
                                       "nxp-demo-experience-assets/releases/download/5.10.52_2.1.0/"
                                       "eIQVideoSwitchCore.zip"},
                     'window_title': "PyeIQ - Object Detection Switch Cores"}

CPU_IMG = os.path.join("/tmp", "cpu.jpg")
GPU_IMG = os.path.join("/tmp", "gpu.jpg")
NPU_IMG = os.path.join("/tmp", "npu.jpg")

JPEG_EOF = b'\xff\xd9'

CPU = 0
GPU = 1
NPU = 2

PAUSE = "kill -STOP {}"
RESUME = "kill -CONT {}"
RUN = "USE_GPU_INFERENCE={} {} -a {}"



#!/usr/bin/env python3
# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os
import pathlib
from random import randint
import shutil
from subprocess import Popen

from eiq import __version__ as version
from eiq.apps.pyeiq_launcher.config import APPS, DEMOS
from eiq.config import BASE_DIR
from eiq.utils import args_parser, colored


class PyeIQ:
    def __init__(self):
        self.args = args_parser()
        self.apps = [key for key in APPS.keys()]
        self.demos = [key for key in DEMOS.keys()]
        self.pyeiq_set = {**APPS, **DEMOS}

    def pyeiq_info(self):
        msg = "Welcome to PyeIQ - v{}".format(version)
        self.main_msg(msg)

        print("Available commands:\n")
        print("# pyeiq --clear-cache: Clear cached data generated by demos.")
        print("# pyeiq --info app/demo: app/demo short description and usage."
              "(e.g., pyeiq --run {}).".format(self.randomize(self.demos)))
        print("# pyeiq --list-apps: List the available applications.")
        print("# pyeiq --list-demos: List the available demos.")
        print("# pyeiq --run app/demo: Run the app/demo "
              "(e.g., pyeiq --run {}).\n".format(self.randomize(self.demos)))

    @staticmethod
    def main_msg(msg):
        os.system("clear")
        msg = "   {}   ".format(msg)
        print(colored((len(msg)+4) * "#", "blue"))
        print(colored("#" + (len(msg)+2) * " " + "#", "blue"))
        print(colored("#", "blue"), colored(msg, "green"), colored("#", "blue"))
        print(colored("#" + (len(msg)+2) * " " + "#", "blue"))
        print(colored((len(msg)+4) * "#" + "\n", "blue"))

    def randomize(self, arr):
        return arr[randint(0, (len(arr) - 1))]

    def clear_cache(self):
        if os.path.exists(BASE_DIR):
            print("Removing {}...".format(BASE_DIR))
            try:
                shutil.rmtree(BASE_DIR)
                print("{} has been removed.".format(BASE_DIR))
            except:
                print("Failed to remove {}.".format(BASE_DIR))
            print("")
        else:
            print(colored("No data to be removed.\n", "yellow"))

    def install(self):
        if os.path.isfile(self.args.install):
            print(f"Installing {self.args.install}...")
            pathlib.Path(BASE_DIR).mkdir(parents=True, exist_ok=True)
            file_name = os.path.join(BASE_DIR, os.path.basename(self.args.install))
            shutil.move(self.args.install, file_name)
            shutil.unpack_archive(file_name, BASE_DIR)
            os.remove(file_name)

            for file in os.listdir(BASE_DIR):
                name = file.split(".")[0]
                path = os.path.join(BASE_DIR, name)
                new_file = os.path.join(path, file)
                pathlib.Path(path).mkdir(parents=True, exist_ok=True)
                shutil.move(os.path.join(BASE_DIR, file), new_file)
                shutil.unpack_archive(new_file, path)

    def print_info(self, target):
        if target in self.pyeiq_set:
            msg = "PyeIQ - {}".format(target)
            self.main_msg(msg)
            print(self.pyeiq_set[target]().description())

            if target in self.demos:
                self.pyeiq_set[target]().usage(target)
        else:
            print(colored("Invalid PyeIQ app/demo.\n"
                          "Type pyeiq --list-demos or --list-apps to "
                          "list the available apps/demos", "yellow"))
        print("")

    def list_apps(self):
        self.main_msg("PyeIQ - Available Applications")

        for app in self.apps:
            print(">>> {}".format(app))

        print("\nFor more details about an application use --info (e.g.,"
              "pyeiq --info {}).".format(self.randomize(self.apps)))
        print("To run an application use --run (e.g., pyeiq --run "
              "{}).\n".format(self.randomize(self.apps)))

    def list_demos(self):
        self.main_msg("PyeIQ - Available Demos")

        for demo in self.demos:
            print(">>> {}".format(demo))

        print("\nFor more details about a demo use --info (e.g.,"
              " pyeiq --info {}).".format(self.randomize(self.demos)))
        print("To run a demo use --run (e.g., pyeiq --run "
              "{}).\n".format(self.randomize(self.demos)))

    def run(self, target):
            if target in self.pyeiq_set:
                msg = "PyeIQ - {}".format(target)
                self.main_msg(msg)
                self.pyeiq_set[target]().run()
            else:
                print(colored("Invalid PyeIQ app/demo.\n"
                              "Type pyeiq --list-demos or --list-apps to "
                              "list the available apps/demos", "yellow"))
            print("")

    def main(self):
        if self.args.install:
            self.install()
        if self.args.clear_cache:
            self.clear_cache()
        elif self.args.info:
            self.print_info(self.args.info)
        elif self.args.list_apps:
            self.list_apps()
        elif self.args.list_demos:
            self.list_demos()
        elif self.args.run:
            self.run(self.args.run)
        else:
            self.pyeiq_info()


if __name__ == '__main__':
    PyeIQ().main()



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import datetime
import os

import cv2


TMP_FILE_PATH = "eiq"
HOME = os.environ['HOME']
BASE_DIR = os.path.join(HOME, ".cache", "eiq")

CHUNK_DEFAULT_SIZE = 32768
REGULAR_DOWNLOAD_URL = 'https://docs.google.com/uc?export=download'

INF_TIME_MSG = "INFERENCE TIME"
FPS_MSG = "FPS"
MODEL_MSG = "MODEL"
QUANT_MSG0 = "WARNING: This model is not full integer quantized!"
QUANT_MSG1 = "Its performance is considerably reduced!"
SRC_MSG = "SOURCE"
MAX_TIME = datetime.timedelta(9, 9, 9)
ID = 5

PNG = ".png"
ZIP = ".zip"

FONT = {'hershey': cv2.FONT_HERSHEY_SIMPLEX,
        'size': 0.8,
        'color': {'black': (0, 0, 0),
                  'blue': (255, 0, 0),
                  'green': (0, 255, 0),
                  'orange': (0, 127, 255),
                  'red': (0, 0, 255),
                  'white': (255, 255, 255)},
        'thickness': 2}



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os
from socket import gethostname
from stat import S_IEXEC
import subprocess
import threading
import time

import cv2
import numpy as np
from PIL import Image

from eiq.apps.switch_video.config import *
from eiq.apps.utils import create_link
from eiq.config import BASE_DIR,ZIP
from eiq.utils import args_parser, check_data, Downloader, file_type


class eIQVideoSwitchCore:
    def __init__(self):
        self.args = args_parser()
        self.pid = [0, 0, 0]
        self.device = None
        self.class_name = self.__class__.__name__
        self.data = VIDEO_SWITCH_CORE

        self.base_dir = os.path.join(BASE_DIR,self.class_name)
        self.binary = os.path.join(self.base_dir, "bin", "video_switch_core")
        self.labels = os.path.join(self.base_dir, "model", "labels.txt")
        self.media = os.path.join(self.base_dir, "media", "video_device.mp4")
        self.media_backup = os.path.join(self.base_dir, "media", "backup.mp4")
        self.model = os.path.join(self.base_dir, "model", "ssd_mobilenet_v1_1_default_1.tflite")
        self.tmp_img = NPU_IMG
        
        self.backup()

    @staticmethod
    def description():
        return ("This demo uses:\n   * TensorFlow Lite as inference engine."
                "\n   * Single Shot Detection as default algorithm.\n\n"
                "This application allows users to run an object detection demo\n"
                "using CPU, GPU or NPU to perform inference on a video.\n")
    
    def gather_data(self):
        downloader = Downloader(self.args)
        downloader.retrieve_data(self.data['src'], self.class_name+ZIP,
                                 self.base_dir, self.data['sha1'], True)
    
    def run_inference(self, device):
        self.pid[device] = subprocess.Popen(RUN.format(device, self.binary, device), shell=True).pid

    def get_device(self):
        hostname = gethostname()

        if "imx8mp" in hostname:
            self.device = "npu"
        elif "imx8" in hostname:
            self.device = "gpu"            

    def pause_proc(self, dev):
        if self.pid[dev] > 0:
            proc = subprocess.Popen(PAUSE.format(self.pid[dev]), shell=True)
            proc.wait()

    def resume_proc(self, dev):
        if self.pid[dev] > 0:
            proc = subprocess.Popen(RESUME.format(self.pid[dev]), shell=True)
            proc.wait()

    def interruption(self):
        self.get_device()

        while True:
            interrupt = str(input("Choose: 'cpu' 'gpu' '{}': ".format(self.device)))
            os.system("clear")

            if interrupt == "cpu":
                self.pause_proc(GPU)
                self.pause_proc(NPU)
                self.resume_proc(CPU)
                self.tmp_img = CPU_IMG
            elif interrupt == "gpu":                
                self.pause_proc(CPU)
                self.pause_proc(NPU)
                self.resume_proc(GPU)
                self.tmp_img = GPU_IMG
            elif interrupt == "npu":                
                self.pause_proc(CPU)
                self.pause_proc(GPU)
                self.resume_proc(NPU)
                self.tmp_img = NPU_IMG
            else:
                print("Invalid option. Please, choose between 'cpu' 'gpu' '{}': ".format(self.device))

    def start_threads(self):
        b = np.zeros([480, 640, 3], 'uint8')
        b[:,:,0] = 255
        b[:,:,1] = 255
        b[:,:,2] = 255
        img = Image.fromarray(b)
        cv_img = cv2.cvtColor(np.array(img), cv2.COLOR_RGB2BGR)
        cv2.imwrite(CPU_IMG, cv_img)
        cv2.imwrite(NPU_IMG, cv_img)
        cv2.imwrite(GPU_IMG, cv_img)

        input_thread = threading.Thread(target=self.interruption)
        
        cpu_inference_thread = threading.Thread(target=self.run_inference,
                                                args=(CPU, ))
        npu_inference_thread = threading.Thread(target=self.run_inference,
                                                args=(GPU, ))
        gpu_inference_thread = threading.Thread(target=self.run_inference,
                                              args=(NPU, ))
                                              
        
        input_thread.daemon = True
        cpu_inference_thread.daemon = True
        npu_inference_thread.daemon = True
        gpu_inference_thread.daemon = True
        npu_inference_thread.start()
        time.sleep(5)
        gpu_inference_thread.start()
        time.sleep(5)
        cpu_inference_thread.start()
        time.sleep(3)
        input_thread.start()

    def backup(self):
        if os.path.isfile(self.media_backup):
            os.system(f"cp {self.media_backup} {self.media}")

    def start(self):
        os.environ['VSI_NN_LOG_LEVEL'] = "0"
        if not check_data(os.path.join(self.base_dir, self.class_name + ZIP),
                          self.data['sha1'], self.binary, self.labels,
                          self.media, self.model):
                          
            self.gather_data()
        os.chmod(self.binary, S_IEXEC)
        create_link()
        
        if self.args.video_src and os.path.isfile(self.args.video_src):
            if file_type(self.args.video_src) != "video":
                print("Your video_src is not a valid video file. Using the default video...")
            else:
                os.system(f"cp {self.media} {self.media_backup}")
                os.system(f"cp {self.args.video_src} {self.media}")
        else:
            print("Your video_src is not a valid video file. Using the default video...")

        self.start_threads()

    def run(self):
        self.start()

        while True:
            with open(self.tmp_img, 'rb') as f:
                bytes_img = f.read()

            if bytes_img.endswith(JPEG_EOF):
                image = cv2.imdecode(np.frombuffer(bytes_img, dtype=np.uint8),
                                     cv2.IMREAD_COLOR)
                cv2.imshow("window_title", image)
                cv2.waitKey(5)



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os
import numpy as np
import pyarmnn as ann
from eiq.utils import Timer


class aNNInterpreter:
    def __init__(self, model=None):
        self.interpreter = None
        self.input_details = None
        self.output_details = None
        self.inference_time = None

        if model is not None:
            parser = ann.ITfLiteParser()
            network = parser.CreateNetworkFromBinaryFile(model)

            graph_id = 0
            input_names = parser.GetSubgraphInputTensorNames(graph_id)
            self.input_binding_info = parser.GetNetworkInputBindingInfo(
                graph_id, input_names[0])

            options = ann.CreationOptions()
            self.runtime = ann.IRuntime(options)

            preferredBackends = [ann.BackendId('VsiNpu'), ann.BackendId('CpuAcc'), ann.BackendId('CpuRef')]
            opt_network, messages = ann.Optimize(
                network, preferredBackends, self.runtime.GetDeviceSpec(), ann.OptimizerOptions())

            net_id, _ = self.runtime.LoadNetwork(opt_network)

            output_names = parser.GetSubgraphOutputTensorNames(graph_id)
            self.output_binding_info = parser.GetNetworkOutputBindingInfo(
                0, output_names[0])
            self.output_tensors = ann.make_output_tensors([self.output_binding_info])
        
    def set_tensor(self, image):
        self.input_tensors = ann.make_input_tensors([self.input_binding_info], [image])

    def get_tensor(self, squeeze=False):
        if squeeze:
            return np.squeeze(ann.workload_tensors_to_ndarray(self.output_tensors))

        return ann.workload_tensors_to_ndarray(self.output_tensors)

    def run_inference(self):
        timer = Timer()
        with timer.timeit():
            self.runtime.EnqueueWorkload(0, self.input_tensors, self.output_tensors)
        self.inference_time = timer.time



# Copyright 2018 The TensorFlow Authors
#
## Copyright 2021 NXP
##
## This file was copied from TensorFlow respecting its rights. All the modified
## parts below are according to TensorFlow's LICENSE terms.
##
## SPDX-License-Identifier:    Apache-2.0

import cv2
import numpy as np
from PIL import Image

from eiq.config import FONT
from eiq.engines.armnn.inference import aNNInterpreter
from eiq.modules.classification.config import OBJ_CLASSIFICATION
from eiq.modules.utils import DemoBase

class eIQObjectClassificationArmNN(DemoBase):
    def __init__(self, args=None):
        super().__init__(args, self.__class__.__name__, OBJ_CLASSIFICATION)

    @staticmethod
    def description():
        return ("This demo uses:\n   * ArmNN as inference engine."
                "\n   * CNN as default algorithm.\n")

    def usage(self, name=None, labels=True, model=True):
        super().usage(name=name, labels=labels, model=model)

    def load_labels(self, label_path):
        with open(label_path, 'r') as f:
            return [line.strip() for line in f.readlines()]

    def process_image(self, image, k=3):
        input_data = np.expand_dims(image, axis=0)
        self.interpreter.set_tensor(input_data)
        self.interpreter.run_inference()
        output_data = self.interpreter.get_tensor(0, squeeze=True)

        top_k = output_data.argsort()[-k:][::-1]
        result = []
        for i in top_k:
            score = float(output_data[i] / 255.0)
            result.append((i, score))
        return result

    def display_result(self, top_result, frame, labels):
        for idx, (i, score) in enumerate(top_result):
            x = 3
            y = 35 * idx + 60
            cv2.putText(frame, '{} - {:0.4f}'.format(labels[i], score),
                        (x, y), FONT['hershey'], FONT['size'],
                        FONT['color']['black'], FONT['thickness'] + 2)
            cv2.putText(frame, '{} - {:0.4f}'.format(labels[i], score),
                        (x, y), FONT['hershey'], FONT['size'],
                        FONT['color']['orange'], FONT['thickness'])

        self.overlay.draw_info(frame, self.model, self.media_src,
                               self.interpreter.inference_time)

    def classify_image(self, frame):
        image = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        image = image.resize((128, 128))

        top_result = self.process_image(image)
        self.display_result(top_result, frame, self.labels)

        return frame

    def start(self):
        self.gather_data()
        self.interpreter = aNNInterpreter(self.model)
        self.labels = self.load_labels(self.labels)

    def run(self):
        self.start()
        self.run_inference(self.classify_image)



# Copyright 2018 The TensorFlow Authors
# Copyright 2021 NXP
#
# This file was copied from TensorFlow respecting its rights. All the modified
# parts below are according to TensorFlow's LICENSE terms.
#
# SPDX-License-Identifier:    Apache-2.0

import collections
import colorsys
import os
import random
import sys
import time

import gi
gi.require_version('Gst', '1.0')
from gi.repository import Gst

import cv2
import numpy as np
from PIL import Image

from eiq.config import FONT
from eiq.engines.opencv.inference import OpenCVDNN
from eiq.engines.tflite.inference import TFLiteInterpreter
from eiq.modules.detection.config import OBJ_DETECTION, OBJ_DETECTION_GST
from eiq.modules.utils import DemoBase
from eiq.multimedia import gstreamer

try:
    import svgwrite
    has_svgwrite = True
except ImportError:
    has_svgwrite = False


class eIQObjectDetection(DemoBase):
    def __init__(self, args=None):
        super().__init__(args, self.__class__.__name__, OBJ_DETECTION)

        self.colors = None

    @staticmethod
    def description():
        return ("This demo uses:\n   * TensorFlow Lite as inference engine."
                "\n   * Single Shot Detection as default algorithm.\n")

    def usage(self, name=None, labels=True, model=True):
        super().usage(name=name, labels=labels, model=model)

    def generate_colors(self):
        hsv_tuples = [(x / len(self.labels), 1., 1.)
                      for x in range(len(self.labels))]

        colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
        colors = list(map(lambda x: (int(x[0] * 255), int(x[1] * 255),
                                     int(x[2] * 255)), colors))
        random.seed(10101)
        random.shuffle(colors)
        random.seed(None)

        self.colors = colors

    def process_image(self, image):
        self.interpreter.set_tensor(np.expand_dims(image, axis=0))
        self.interpreter.run_inference()

        positions = self.interpreter.get_tensor(0, squeeze=True)
        classes = self.interpreter.get_tensor(1, squeeze=True)
        scores = self.interpreter.get_tensor(2, squeeze=True)

        result = []
        for idx, score in enumerate(scores):
            if score > 0.5:
                result.append({'pos': positions[idx], '_id': classes[idx]})
        return result

    def display_result(self, frame, result):
        width = frame.shape[1]
        height = frame.shape[0]

        for obj in result:
            pos = obj['pos']
            _id = obj['_id']

            x1 = int(pos[1] * width)
            x2 = int(pos[3] * width)
            y1 = int(pos[0] * height)
            y2 = int(pos[2] * height)

            top = max(0, np.floor(y1 + 0.5).astype('int32'))
            left = max(0, np.floor(x1 + 0.5).astype('int32'))
            bottom = min(height, np.floor(y2 + 0.5).astype('int32'))
            right = min(width, np.floor(x2 + 0.5).astype('int32'))

            label_size = cv2.getTextSize(self.labels[_id], FONT['hershey'],
                                         FONT['size'], FONT['thickness'])[0]
            label_rect_left = int(left - 3)
            label_rect_top = int(top - 3)
            label_rect_right = int(left + 3 + label_size[0])
            label_rect_bottom = int(top - 5 - label_size[1])

            cv2.rectangle(frame, (left, top), (right, bottom),
                          self.colors[int(_id) % len(self.colors)], 6)
            cv2.rectangle(frame, (label_rect_left, label_rect_top),
                          (label_rect_right, label_rect_bottom),
                          self.colors[int(_id) % len(self.colors)], -1)
            cv2.putText(frame, self.labels[_id], (left, int(top - 4)),
                        FONT['hershey'], FONT['size'],
                        FONT['color']['black'],
                        FONT['thickness'])
            self.overlay.draw_info(frame, self.model, self.media_src,
                                   self.interpreter.inference_time,
                                   self.interpreter.input_details[0]['quantization'])

    def detect_object(self, frame):
        image = Image.fromarray(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
        image = image.resize((self.interpreter.width(), self.interpreter.height()))
        top_result = self.process_image(image)
        self.display_result(frame, top_result)

        return frame

    def start(self):
        self.gather_data()
        self.validate_data(self.image, self.labels, self.model)
        self.interpreter = TFLiteInterpreter(self.model)
        self.labels = self.load_labels(self.labels)
        self.generate_colors()

    def run(self):
        self.start()
        self.run_inference(self.detect_object)


class eIQObjectDetectionGStreamer(DemoBase):
    def __init__(self, args=None):
        super().__init__(args, self.__class__.__name__, OBJ_DETECTION_GST)

        self.tensor = None

        self.videosrc = None
        self.videofile = None
        self.videofmt = "raw"
        self.src_width = 640
        self.src_height = 480

    def video_config(self):
        if self.args.video_src and self.args.video_src.startswith("/dev/video"):
            self.videosrc = self.args.video_src
        elif self.args.video_src and os.path.exists(self.args.video_src):
            self.videofile = self.args.video_src
            self.src_width = 1920
            self.src_height = 1080

    def input_image_size(self):
        return self.interpreter.input_details[0]['shape'][1:]

    def input_tensor(self):
        return self.tensor(self.interpreter.input_details[0]['index'])()[0]

    def set_input(self, buf):
        result, mapinfo = buf.map(Gst.MapFlags.READ)
        if result:
            np_buffer = np.reshape(np.frombuffer(mapinfo.data, dtype=np.uint8),
                                   self.input_image_size())
            self.input_tensor()[:, :] = np_buffer
            buf.unmap(mapinfo)

    def output_tensor(self, i):
        output_data = np.squeeze(self.tensor(
                                 self.interpreter.output_details[i]['index'])())
        if 'quantization' not in self.interpreter.output_details:
            return output_data
        scale, zero_point = self.interpreter.output_details['quantization']
        if scale == 0:
            return output_data - zero_point
        return scale * (output_data - zero_point)

    @staticmethod
    def avg_fps_counter(window_size):
        window = collections.deque(maxlen=window_size)
        prev = time.monotonic()
        yield 0.0

        while True:
            curr = time.monotonic()
            window.append(curr - prev)
            prev = curr
            yield len(window) / sum(window)

    @staticmethod
    def shadow_text(dwg, x, y, text, font_size=20):
        dwg.add(dwg.text(text, insert=(x + 1, y + 1),
                         fill='black', font_size=font_size))
        dwg.add(dwg.text(text, insert=(x, y),
                fill='white', font_size=font_size))

    def generate_svg(self, src_size, inference_size,
                     inference_box, objs, labels, text_lines):
        dwg = svgwrite.Drawing('', size=src_size)
        src_w, src_h = src_size
        inf_w, inf_h = inference_size
        box_x, box_y, box_w, box_h = inference_box
        scale_x, scale_y = src_w / box_w, src_h / box_h

        for y, line in enumerate(text_lines, start=1):
            self.shadow_text(dwg, 10, y * 20, line)
        for obj in objs:
            x0, y0, x1, y1 = list(obj.bbox)
            x, y, w, h = x0, y0, x1 - x0, y1 - y0
            x, y, w, h = int(x * inf_w), int(y * inf_h),\
                         int(w * inf_w), int(h * inf_h)
            x, y = x - box_x, y - box_y
            x, y, w, h = x * scale_x, y * scale_y, w * scale_x, h * scale_y
            percent = int(100 * obj.score)
            label = '{}% {}'.format(percent, labels.get(obj.id, obj.id))
            self.shadow_text(dwg, x, y - 5, label)
            dwg.add(dwg.rect(insert=(x, y), size=(w, h), fill='none',
                             stroke='red', stroke_width='2'))
        return dwg.tostring()

    def get_output(self, score_threshold=0.1, top_k=3):
        boxes = self.output_tensor(0)
        category = self.output_tensor(1)
        scores = self.output_tensor(2)

        return [gstreamer.make_boxes(i, boxes, category, scores)
                for i in range(top_k) if scores[i] >= score_threshold]

    def start(self):
        self.video_config()
        self.gather_data()
        self.interpreter = TFLiteInterpreter(self.model)
        self.tensor = self.interpreter.interpreter.tensor

    def run(self):
        if not has_svgwrite:
            sys.exit("The module svgwrite needed to run this demo was not "
                     "found. If you want to install it type 'pip3 install "
                     "svgwrite' at your terminal. Exiting...")

        self.start()
        labels = self.load_labels(self.labels)
        w, h, _ = self.input_image_size()
        inference_size = (w, h)
        fps_counter = self.avg_fps_counter(30)

        def user_callback(input_tensor, src_size, inference_box):
            nonlocal fps_counter
            start_time = time.monotonic()
            self.set_input(input_tensor)
            self.interpreter.run_inference()
            objs = self.get_output()
            end_time = time.monotonic()
            text_lines = ['Inference: {:.2f} ms'.format((end_time-start_time)
                                                        * 1000),
                          'FPS: {} fps'.format(round(next(fps_counter)))]
            return self.generate_svg(src_size, inference_size, inference_box,
                                     objs, labels, text_lines)

        gstreamer.run_pipeline(user_callback,
                               src_size=(self.src_width,
                                         self.src_height),
                               appsink_size=inference_size,
                               videosrc=self.videosrc,
                               videofile=self.videofile,
                               videofmt=self.videofmt)



# Copyright 2019 Google LLC
# Copyright 2021 NXP
#
# This file was copied from Google LLC respecting its rights. All the modified
# parts below are according to Google LLC's LICENSE terms.
#
# SPDX-License-Identifier:    Apache-2.0

import collections
import sys

try:
    import svgwrite
except ImportError:
    None

import threading

import gi
gi.require_version('Gst', '1.0')
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gst, Gtk

import numpy as np


Object = collections.namedtuple('Object', ['id', 'score', 'bbox'])

GObject.threads_init()
Gst.init(None)


class BBox(collections.namedtuple('BBox', ['xmin', 'ymin', 'xmax', 'ymax'])):
    __slots__ = ()


def make_boxes(i, boxes, class_ids, scores):
    ymin, xmin, ymax, xmax = boxes[i]
    return Object(id=int(class_ids[i]), score=scores[i],
                  bbox=BBox(xmin=np.maximum(0.0, xmin),
                            ymin=np.maximum(0.0, ymin),
                            xmax=np.minimum(1.0, xmax),
                            ymax=np.minimum(1.0, ymax)))

def set_appsink_video_pipeline(device,
                        leaky="leaky=downstream max-size-buffers=1",
                        sync="drop=True max-buffers=1 emit-signals=True max-lateness=8000000000"):

    return (("filesrc location={} ! qtdemux name=d d.video_0 ! \
                queue ! decodebin ! videorate ! videoconvert ! \
                videoscale n-threads=4 method=nearest-neighbour ! \
                video/x-raw,format=RGB,width=1920,height=1080 ! \
                queue {} ! appsink name=sink {}"
            ).format(device,
                    leaky, sync))

def set_appsink_pipeline(device,
                        leaky="leaky=downstream max-size-buffers=1",
                        sync="drop=True max-buffers=1 emit-signals=True max-lateness=8000000000"):

    return (("v4l2src device={} do-timestamp=True ! videoconvert ! \
                videoscale n-threads=4 method=nearest-neighbour ! \
                video/x-raw,format=RGB,width=1920,height=1080 ! \
                queue {} ! appsink name=sink {}"
            ).format(device,
                    leaky, sync))

def set_appsrc_pipeline(width=1920, height=1080):
    return (("appsrc name=src is-live=True block=True ! \
                video/x-raw,format=RGB,width=1920,height=1080, \
                framerate=30/1,interlace-mode=(string)progressive ! \
                videoconvert ! imxvideoconvert_g2d ! waylandsink"
            ).format(width, height))

class GstPipeline:
    def __init__(self, pipeline, user_function, src_size):
        self.user_function = user_function
        self.running = False
        self.gstbuffer = None
        self.sink_size = None
        self.src_size = src_size
        self.box = None
        self.condition = threading.Condition()

        self.pipeline = Gst.parse_launch(pipeline)
        self.overlay = self.pipeline.get_by_name('overlay')
        self.overlaysink = self.pipeline.get_by_name('overlaysink')
        appsink = self.pipeline.get_by_name('appsink')
        appsink.connect('new-sample', self.on_new_sample)

        bus = self.pipeline.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_bus_message)

    def run(self):
        self.running = True
        worker = threading.Thread(target=self.inference_loop)
        worker.start()

        self.pipeline.set_state(Gst.State.PLAYING)
        try:
            Gtk.main()
        except BaseException:
            pass

        self.pipeline.set_state(Gst.State.NULL)
        while GLib.MainContext.default().iteration(False):
            pass
        with self.condition:
            self.running = False
            self.condition.notify_all()
        worker.join()

    def on_bus_message(self, bus, message):
        t = message.type
        if t == Gst.MessageType.EOS:
            Gtk.main_quit()
        elif t == Gst.MessageType.WARNING:
            err, debug = message.parse_warning()
            sys.stderr.write('Warning: %s: %s\n' % (err, debug))
        elif t == Gst.MessageType.ERROR:
            err, debug = message.parse_error()
            sys.stderr.write('Error: %s: %s\n' % (err, debug))
            Gtk.main_quit()
        return True

    def on_new_sample(self, sink):
        sample = sink.emit('pull-sample')
        if not self.sink_size:
            s = sample.get_caps().get_structure(0)
            self.sink_size = (s.get_value('width'), s.get_value('height'))
        with self.condition:
            self.gstbuffer = sample.get_buffer()
            self.condition.notify_all()
        return Gst.FlowReturn.OK

    def get_box(self):
        if not self.box:
            glbox = self.pipeline.get_by_name('glbox')
            if glbox:
                glbox = glbox.get_by_name('filter')
            box = self.pipeline.get_by_name('box')
            assert glbox or box
            assert self.sink_size
            if glbox:
                self.box = (glbox.get_property('x'), glbox.get_property('y'),
                            glbox.get_property('width'), glbox.get_property('height'))
            else:
                self.box = (-box.get_property('left'), -box.get_property('top'),
                            self.sink_size[0] + box.get_property(
                                'left') + box.get_property('right'),
                            self.sink_size[1] + box.get_property('top') + box.get_property('bottom'))
        return self.box

    def inference_loop(self):
        while True:
            with self.condition:
                while not self.gstbuffer and self.running:
                    self.condition.wait()
                if not self.running:
                    break
                gstbuffer = self.gstbuffer
                self.gstbuffer = None

            input_tensor = gstbuffer
            svg = self.user_function(
                input_tensor, self.src_size, self.get_box())
            if svg:
                if self.overlay:
                    self.overlay.set_property('data', svg)
                if self.overlaysink:
                    self.overlaysink.set_property('svg', svg)

def camera_pipeline(video_src):
    PIPELINE = 'v4l2src device=%s ! {src_caps}' % video_src
    PIPELINE += """ ! tee name=t
        t. ! {leaky_q} ! imxvideoconvert_g2d ! {scale_caps} ! videobox name=box autocrop=true
           ! {sink_caps} ! {sink_element}
        t. ! queue ! imxvideoconvert_g2d
           ! rsvgoverlay name=overlay ! waylandsink
        """
    return PIPELINE

def video_file_pipeline(video_src):
    PIPELINE = 'filesrc location=%s ! decodebin' % video_src
    PIPELINE += """ ! tee name=t
        t. ! {leaky_q} ! imxvideoconvert_g2d ! {scale_caps} ! videobox name=box autocrop=true
           ! {sink_caps} ! {sink_element}
        t. ! queue ! imxvideoconvert_g2d
           ! rsvgoverlay name=overlay ! waylandsink
    """
    return PIPELINE

def run_pipeline(user_function,
                 src_size,
                 appsink_size,
                 videosrc='/dev/video1',
                 videofile='None',
                 videofmt='raw'):
    scale = min(
        appsink_size[0] /
        src_size[0],
        appsink_size[1] /
        src_size[1])
    scale = tuple(int(x * scale) for x in src_size)
    scale_caps = 'video/x-raw,width={width},height={height}'.format(
        width=scale[0], height=scale[1])
    if videofile != None:
        PIPELINE = video_file_pipeline(videofile)

    else:
        PIPELINE = camera_pipeline(videosrc)

    SRC_CAPS = 'video/x-raw,width={width},height={height},framerate=30/1'
    SINK_ELEMENT = 'appsink name=appsink emit-signals=true max-buffers=1 drop=true'
    SINK_CAPS = 'video/x-raw,format=RGB,width={width},height={height}'
    LEAKY_Q = 'queue max-size-buffers=1 leaky=downstream'

    src_caps = SRC_CAPS.format(width=src_size[0], height=src_size[1])
    sink_caps = SINK_CAPS.format(width=appsink_size[0], height=appsink_size[1])
    pipeline = PIPELINE.format(leaky_q=LEAKY_Q,
                               src_caps=src_caps, sink_caps=sink_caps,
                               sink_element=SINK_ELEMENT, scale_caps=scale_caps)

    print('Gstreamer pipeline:\n', pipeline)

    pipeline = GstPipeline(pipeline, user_function, src_size)
    pipeline.run()



# Copyright 2017 Andrea Palazzi 
# Copyright 2021 NXP
#
# This Google Driver class was copied from Andrea Palazzi respecting its rights.
# All the modified parts below are according to MIT's LICENSE terms.
#
# SPDX-License-Identifier: MIT
# Reference:
# https://github.com/ndrplz/google-drive-downloader/blob/master/LICENSE

from __future__ import print_function
import zipfile
import warnings
from sys import stdout
from os import makedirs
from os.path import dirname
from os.path import exists

import requests

from eiq import config

class GoogleDriveDownloader:
    CHUNK_SIZE = config.CHUNK_DEFAULT_SIZE
    DOWNLOAD_URL = config.REGULAR_DOWNLOAD_URL

    @staticmethod
    def download_file_from_google_drive(
            file_id, dest_path, overwrite=False, unzip=False, showsize=False):
        destination_directory = dirname(dest_path)
        if not exists(destination_directory):
            makedirs(destination_directory)

        if not exists(dest_path) or overwrite:
            session = requests.Session()
            print("Downloading {} into {}... ".format(
                file_id, dest_path), end='')
            stdout.flush()
            response = session.get(
                GoogleDriveDownloader.DOWNLOAD_URL, params={
                    'id': file_id}, stream=True)

            token = GoogleDriveDownloader._get_confirm_token(response)
            if token:
                params = {'id': file_id, 'confirm': token}
                response = session.get(
                    GoogleDriveDownloader.DOWNLOAD_URL,
                    params=params,
                    stream=True)
            if showsize: print()

            current_download_size = [0]
            GoogleDriveDownloader._save_response_content(
                response, dest_path, showsize, current_download_size)
            print("done.")

            if unzip:
                try:
                    print("Unzipping...", end='')
                    stdout.flush()
                    with zipfile.ZipFile(dest_path, 'r') as z:
                        z.extractall(destination_directory)
                    print("done.")
                except zipfile.BadZipfile:
                    warnings.warn(
                        "Ignoring unzip since '{}' does not " \
                        "look like a valid zip file".format(file_id))

    @staticmethod
    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith("download_warning"):
                return value
        return None

    @staticmethod
    def _save_response_content(response, destination, showsize, current_size):
        with open(destination, 'wb') as f:
            for chunk in response.iter_content(
                    GoogleDriveDownloader.CHUNK_SIZE):
                if chunk:
                    f.write(chunk)
                    if showsize:
                        print("\r" + GoogleDriveDownloader.sizeof_fmt(
                                current_size[0]), end=' ')
                        stdout.flush()
                        current_size[0] += GoogleDriveDownloader.CHUNK_SIZE

    @staticmethod
    def sizeof_fmt(num, suffix='B'):
        for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
            if abs(num) < 1024.0:
                return "{:.1f} {}{}".format(num, unit, suffix)
            num /= 1024.0
        return "{:.1f} {}{}".format(num, 'Yi', suffix)



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os

import numpy as np
from tflite_runtime.interpreter import Interpreter

from eiq.utils import Timer


class TFLiteInterpreter:
    def __init__(self, model=None):
        os.environ['VSI_NN_LOG_LEVEL'] = "0"
        self.interpreter = None
        self.input_details = None
        self.output_details = None
        self.inference_time = None

        if model is not None:
            self.interpreter = Interpreter(model)
            self.interpreter.allocate_tensors()

            self.input_details = self.interpreter.get_input_details()
            self.output_details = self.interpreter.get_output_details()

    def dtype(self):
        return self.input_details[0]['dtype']

    def height(self):
        return self.input_details[0]['shape'][1]

    def width(self):
        return self.input_details[0]['shape'][2]

    def get_tensor(self, index, squeeze=False):
        if squeeze:
            return np.squeeze(self.interpreter.get_tensor(
                                   self.output_details[index]['index']))

        return self.interpreter.get_tensor(
                    self.output_details[index]['index'])

    def set_tensor(self, image):
        self.interpreter.set_tensor(self.input_details[0]['index'], image)

    def run_inference(self):
        timer = Timer()
        with timer.timeit():
            self.interpreter.invoke()
        self.inference_time = timer.time



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os
from setuptools import setup, find_packages

PYEIQ_LAUNCHER = os.path.join(os.getcwd(), "eiq", "apps", "pyeiq_launcher", "pyeiq.py")
PYEIQ_USR = os.path.join("/usr", "bin", "pyeiq")

if os.path.exists(PYEIQ_USR):
    os.system("rm -rf {}".format(PYEIQ_USR))

os.system("cp {} {}".format(PYEIQ_LAUNCHER, PYEIQ_USR))

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name="pyeiq",
      version="3.0.1",
      description="A Python Framework for eIQ on i.MX Processors",
      long_description=long_description,
      long_description_content_type='text/markdown',
      url = 'https://source.codeaurora.org/external/imxsupport/pyeiq/',
      author="Alifer Moraes, Diego Dorta, Marco Franchi, Kaixin Ding, Gnar Fang, Michael Pontikes, Jinghui Liu, Manish Bajaj, Devin Jiao",
      license="BSD-3-Clause",
      packages=find_packages(),
      zip_safe=False,
      install_requires=['requests'],
      keywords = ['ml', 'eiq', 'demos', 'apps'],
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Information Technology',
        'Natural Language :: English',
        'Operating System :: Other OS',
        'Programming Language :: Python :: 3.9'
      ])



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

# Switch Label Image

LIB_PATH = "/usr/lib"
LIBNN = "libneuralnetworks.so"

SWITCH_IMAGE = {'bin': "label_image",
                'labels': "labels.txt",
                'model': "mobilenet_v1_1.0_224_quant.tflite",
                'sha1': "6cd4ca55c0beb56bcd7f053d2f37cc612f58c710",
                'src': {'drive': "https://drive.google.com/file/d/"
                                 "1omIHVWVR06evXYG1pLII1kodElFujfFR",
                        'github': "https://github.com/KaixinDing/"
                                  "pyeiq_model/releases/download/3.0.0/"
                                  "eIQSwitchLabelImage.zip"},
                'msg': {'confidence': "<b>CONFIDENCE</b>",
                        'inf_time': "<b>INFERENCE TIME</b>",
                        'labels': "<b>LABELS</b>",
                        'model_name': "<b>MODEL NAME</b>",
                        'select_img': "<b>SELECT AN IMAGE</b>"},
                'window_title': "PyeIQ - Label Image Switch App"}

RUN_LABEL_IMAGE = "VSI_NN_LOG_LEVEL=0 {0} -m {1} -t 1 -i {2} -l {3} -a {4} -v 0 -c 100"

REGEX_GET_INTEGER_FLOAT = "\d+\.\d+|\d+"
REGEX_GET_STRING = "[^a-zA-Z\s]"



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

def v4l2_camera_pipeline(width, height, device, frate,
                         leaky="leaky=downstream max-size-buffers=1"):

    return (("v4l2src device={} ! video/x-raw,width={},height={},framerate={} "\
             "! queue {} ! videoconvert ! appsink").format(device, width,
                                                              height, frate,
                                                              leaky))


def v4l2_video_pipeline(device, leaky="leaky=downstream max-size-buffers=1"):

    return (("filesrc location={} ! qtdemux name=d d.video_0 ! decodebin ! queue {} ! queue ! imxvideoconvert_g2d ! videoconvert ! "\
             "appsink").format(device, leaky))



# Copyright 2021 NXP
# SPDX-License-Identifier: BSD-3-Clause

import os
import re
import sys
import threading

import cv2

import gi
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk

from eiq.config import BASE_DIR, ZIP
from eiq.multimedia.overlay import OpenCVOverlay
from eiq.multimedia.utils import GstVideo, save_image, VideoConfig
from eiq.utils import args_parser, check_data, colored, Downloader, file_type, Framerate


class DemoBase:
    def __init__(self, args=None, class_name=None, data=None):
        if args:
            self. args = args
        else:
            self.args = args_parser()

        self.class_name = class_name
        self.framerate = Framerate()
        self.overlay = OpenCVOverlay()

        self.base_dir = os.path.join(BASE_DIR, self.class_name)
        self.save_dir = os.path.join(BASE_DIR, "media")
        self.media_dir = os.path.join(self.base_dir, "media")
        self.model_dir = os.path.join(self.base_dir, "model")

        self.data = data
        self.image = None
        self.labels = None
        self.media_src = None
        self.model = None

        self.interpreter = None

    def usage(self, name=None, labels=False, model=False):
        print("Available parameters:\n")
        print(colored("-d / --download", "yellow") + ": Chooses from which"
              " server the models are going to be downloaded.\n"
              "Options: drive, github, wget. Default: Finds the fastest one.\n"
              "# pyeiq --run {} --download drive\n".format(name))
        print(colored("-i / --image", "yellow") + ": Run the demo on an image"
              " of your choice.\n"
              "# pyeiq --run {} --image /home/root/image.jpg\n".format(name))
        if labels:
            print(colored("-l / --labels", "yellow")+ ": Uses a labels file"
                  " of your choice within the demo.\n"
                  "Labels and models must be compatible for proper outputs.\n"
                  "# pyeiq --run {} --labels /home/root/labels.txt\n".format(name))
        if model:
            print(colored("-m / --model", "yellow") + ": Uses a model file of"
                  " your choice within the demo.\n"
                  "Labels and models must be compatible for proper outputs.\n"
                  "# pyeiq --run {} --model /home/root/model.tflite\n".format(name))
        print(colored("-r / --res", "yellow") + ": Choose the resolution of"
              " your video capture device.\n"
              "Options: full_hd (1920x1080), hd (1280x720), vga (640x480).\n"
              "Default: hd if supported, otherwise uses the best one available.\n"
              "# pyeiq --run {} --res vga\n".format(name))
        print(colored("-f / --video_fwk", "yellow") + ": Choose which framework"
              " is used to display the video.\n"
              "Options: opencv, v4l2, gstreamer (experimental). Default: v4l2.\n"
              "# pyeiq --run {} --video_fwk opencv\n".format(name))
        print(colored("-v / --video_src", "yellow") + ": Run inference in videos."
              "It can be from a video capture device or a video file.\n"
              "Options: True (Find a video capture device automatically),"
              " /dev/video<x>, path_to_video_file.\n"
              "# pyeiq --run {} --video_src /dev/video3\n".format(name))
        print(colored("-c / --camera_params", "yellow") + ": Only used in demo covid19 detection."
              "Choose eye height, eye angle and eye slope for camera.\n"
              'Options: None or "num num num"\n'
              '# pyeiq --run covid19_detection --camera_params "200 60 60"\n'.format(name))
        print(colored("-w / --work_mode", "yellow") + ": Only used in demo covid19 detection."
              "Choose work mode for demo.\n"
              "Options: 0 (Only detect facemask), 1 (Detect both facemask and social distance)\n"
              "# pyeiq --run covid19_detection --work_mode 1\n".format(name))
        print(colored("Multiple parameters can be used at once:\n", "yellow") +
              "# pyeiq --run {} -d drive -v True -f opencv -res vga\n".format(name))

    def validate_data(self, *args):
        if not check_data(os.path.join(self.base_dir, self.class_name + ZIP),
                                       self.data['sha1'], *args):
            self.gather_data()

    def gather_data(self):
        if self.data and 'src' in self.data:
            downloader = Downloader(self.args)
            downloader.retrieve_data(self.data['src'], self.class_name + ZIP,
                                     self.base_dir, self.data['sha1'], True)

        self.validate_media()

        if hasattr(self.args, 'image'):
            if self.args.image and (file_type(self.args.image) == 'image'):
                self.image = self.args.image
        if not self.image and self.data and 'image' in self.data:
            self.image = os.path.join(self.media_dir, self.data['image'])
        if self.image:
            self.media_src = os.path.basename(self.image)

        if hasattr(self.args, 'labels'):
            if self.args.labels and os.path.isfile(self.args.labels):
                self.labels = self.args.labels
        if not self.labels and self.data and 'labels' in self.data:
            self.labels = os.path.join(self.model_dir, self.data['labels'])

        if hasattr(self.args, 'model'):
            if self.args.model and os.path.isfile(self.args.model):
                self.model = self.args.model
        if not self.model and self.data and 'model' in self.data:
            self.model = os.path.join(self.model_dir, self.data['model'])

    def validate_media(self):
        if self.args.image and (not self.args.video_src):
            media = file_type(self.args.image)
            if media == "video":
                sys.exit("Your image parameter seems to be a video file. "
                         "Try to use -v instead.")
            elif (not media) or (media != "image"):
                print("Your image parameter is not a valid image file. "
                      "Using default image...")

        if self.args.video_src:
            media = file_type(self.args.video_src)
            if media == "image":
                sys.exit("Your video_src parameter seems to be an image file. "
                         "Try to use -i instead.")
            elif media and (media != "video"):
                sys.exit("Your video parameter is not a valid video file.")

    @staticmethod
    def load_labels(path):
        p = re.compile(r'\s*(\d+)(.+)')
        with open(path, 'r', encoding='utf-8') as f:
            lines = (p.match(line).groups() for line in f.readlines())
            return {int(num): text.strip() for num, text in lines}

    def run_inference(self, inference_func):
        if self.args.video_src:
            while True:
                video_config = VideoConfig(self.args)
                sink, src = video_config.get_config()

                if os.path.isfile(self.args.video_src):
                    self.media_src = os.path.basename(self.args.video_src)
                else:
                    self.media_src = video_config.dev.name

                if not src:
                    if (not sink) or (not sink.isOpened()):
                        sys.exit("Your video device could not be initialized. Exiting...")
                    while sink.isOpened():
                        ret, frame = sink.read()
                        if ret:
                            self.overlay.draw_fps(frame, round(self.framerate.fps))
                            with self.framerate.fpsit():
                                cv2.imshow(self.data['window_title'], inference_func(frame))
                        else:
                            print("Your video device could not capture any image.")
                            break
                        if (cv2.waitKey(1) & 0xFF) == ord('q'):
                            break
                    sink.release()
                else:
                    with self.framerate.fpsit():
                        gst_video = GstVideo(sink, src, inference_func, self.framerate.fps)
                        gst_video.run()
        else:
            try:
                print("Running inference...")
                inference_func(cv2.imread(self.image, cv2.IMREAD_COLOR))
                frame = cv2.imread(self.image, cv2.IMREAD_COLOR)
                thread = threading.Thread(target=display_image,
                                          args=(self.data['window_title'],
                                                inference_func(frame),
                                                self.save_dir,
                                                self.class_name))
                thread.daemon = True
                thread.start()
                thread.join()
            except KeyboardInterrupt:
                sys.exit("")

        cv2.destroyAllWindows()


def display_image(window_title, image, dest, name):
    save_image(image, dest, name)
    if Gdk.Display.get_default() is not None:
        if(Gdk.Display.get_default().get_monitor(0) != None):
            cv2.imshow(window_title, image)
            cv2.waitKey()



# Copyright 2021 NXP
#
# This file was partial copied from the link below and modified according to
# its license terms.
#
# Reference:
# https://github.com/toradex/torizon-samples/blob/master/dlr-gstreamer/inference.py
#
# SPDX-License-Identifier: BSD-3-Clause

import atexit
import os
import pathlib
import sys

import gi
gi.require_version('Gst', '1.0')
from gi.repository import Gst

import cv2
import numpy as np

from eiq.config import PNG
from eiq.multimedia.gstreamer import set_appsink_pipeline, set_appsink_video_pipeline, set_appsrc_pipeline
from eiq.multimedia.v4l2 import v4l2_camera_pipeline, v4l2_video_pipeline
from eiq.multimedia.overlay import OpenCVOverlay


class VideoDevice:
    def __init__(self):
        self.name = None
        self.caps = None
        self.default_caps = None
        self.full_hd_caps = None
        self.hd_caps = None
        self.vga_caps = None


class Caps:
    def __init__(self):
        self.name = None
        self.format = None
        self.width = None
        self.height = None
        self.framerate = None


class Devices:
    def __init__(self):
        self.devices = []

    def get_video_devices(self):
        Gst.init()
        dev_monitor = Gst.DeviceMonitor()
        dev_monitor.add_filter("Video/Source")
        dev_monitor.start()

        for dev in dev_monitor.get_devices():
            video_dev = VideoDevice()
            dev_props = dev.get_properties()
            dev_caps = dev.get_caps()

            name = dev_props.get_string("device.path")
            caps = self.get_device_caps(dev_caps.normalize())
            full_hd_caps, hd_caps, vga_caps = self.get_std_caps(caps)
            default_caps = hd_caps
            if (not default_caps) and caps:
                default_caps = caps[0]

            video_dev.name = name
            video_dev.caps = caps
            video_dev.default_caps = default_caps
            video_dev.full_hd_caps = full_hd_caps
            video_dev.hd_caps = hd_caps
            video_dev.vga_caps = vga_caps
            self.devices.append(video_dev)

        dev_monitor.stop()

    @staticmethod
    def get_device_caps(dev_caps):
        caps_list = []

        for i in range(dev_caps.get_size()):
            if dev_caps.get_structure(i).get_name() != "video/x-raw":
                continue

            caps = Caps()
            caps_struct = dev_caps.get_structure(i)
            caps.name = caps_struct.get_name()
            caps.format = caps_struct.get_string("format")
            caps.width = caps_struct.get_int("width")[1]
            caps.height = caps_struct.get_int("height")[1]
            framerate = ("{}/{}".format(*caps_struct.get_fraction(
                                        "framerate")[1:]))
            caps.framerate = framerate
            caps_list.append(caps)

        return caps_list

    @staticmethod
    def get_std_caps(dev_caps):
        full_hd_caps = None
        hd_caps = None
        vga_caps = None

        for caps in dev_caps:
            if  (caps.width == 1920) and (caps.height == 1080):
                full_hd_caps = caps
            elif (caps.width == 1280) and (caps.height == 720):
                hd_caps = caps
            elif (caps.width == 640) and (caps.height == 480):
                vga_caps = caps

        return full_hd_caps, hd_caps, vga_caps

    def search_device(self, dev_name):
        dev = None

        if dev_name.startswith("/dev/video"):
            for device in self.devices:
                if device.name == dev_name:
                    dev = device

                if not dev:
                    print("The specified video_src was not found.\n"
                          "Searching for default video device...")

        if not dev and self.devices:
            dev = self.devices[0]
        elif not dev:
            sys.exit("No video device found. Exiting...")

        print("Using {} as video device".format(dev.name))
        return dev


class GstVideo:
    def __init__(self, sink, src, inference_func, framerate):
        self.sink = sink
        self.src = src
        self.inference_func = inference_func
        self.fps = framerate
        self.appsource = None
        self.sink_pipeline = None
        self.src_pipeline = None
        self.overlay = OpenCVOverlay()

        atexit.register(self.exit_handler)

    def exit_handler(self):
        self.sink_pipeline.set_state(Gst.State.NULL)
        self.src_pipeline.set_state(Gst.State.NULL)

    def run(self):
        self.sink_pipeline = Gst.parse_launch(self.sink)
        appsink = self.sink_pipeline.get_by_name('sink')
        appsink.connect("new-sample", self.on_new_frame)

        self.src_pipeline = Gst.parse_launch(self.src)
        self.appsource = self.src_pipeline.get_by_name('src')

        self.sink_pipeline.set_state(Gst.State.PLAYING)
        bus1 = self.sink_pipeline.get_bus()
        self.src_pipeline.set_state(Gst.State.PLAYING)
        bus2 = self.src_pipeline.get_bus()

        while True:
            message = bus1.timed_pop_filtered(10000, Gst.MessageType.ANY)
            if message:
                if message.type == Gst.MessageType.ERROR:
                    err, debug = message.parse_error()
                    self.sink_pipeline.set_state(Gst.State.NULL)
                    self.src_pipeline.set_state(Gst.State.NULL)
                    sys.exit("ERROR bus 1: {}: {}".format(err, debug))

                if message.type == Gst.MessageType.WARNING:
                    err, debug = message.parse_warning()
                    print("WARNING bus 1: {}: {}".format(err, debug))

            message = bus2.timed_pop_filtered(10000, Gst.MessageType.ANY)
            if message:
                if message.type == Gst.MessageType.ERROR:
                    err, debug = message.parse_error()
                    self.sink_pipeline.set_state(Gst.State.NULL)
                    self.src_pipeline.set_state(Gst.State.NULL)
                    sys.exit("ERROR bus 2: {}: {}".format(err, debug))

                if message.type == Gst.MessageType.WARNING:
                    err, debug = message.parse_warning()
                    print("WARNING bus 2: {}: {}".format(err, debug))

    def on_new_frame(self, sink):
        sample = sink.emit("pull-sample")
        caps = sample.get_caps().get_structure(0)
        resize = (caps.get_value('height'), caps.get_value('width'), 3)

        mem = sample.get_buffer()
        _, arr = mem.map(Gst.MapFlags.READ)
        img = np.ndarray(resize, buffer=arr.data, dtype=np.uint8)

        self.overlay.draw_fps(img, round(self.fps))

        img = self.inference_func(img)
        self.appsource.emit("push-buffer", Gst.Buffer.new_wrapped(img.tobytes()))
        mem.unmap(arr)

        return Gst.FlowReturn.OK


class VideoConfig:
    def __init__(self, args):
        self.res = args.res
        self.video_fwk = args.video_fwk
        self.video_src = args.video_src
        self.devices = Devices()
        self.devices.get_video_devices()

        if not os.path.isfile(self.video_src):
            self.dev = self.devices.search_device(self.video_src)
            self.dev_caps = self.get_caps()

    def get_caps(self):
        if self.res == "full_hd":
                return self.validate_caps(self.dev.full_hd_caps)
        elif self.res == "vga":
                return self.validate_caps(self.dev.vga_caps)
        else:
            if self.res != "hd":
                print("Invalid resolution, trying to use hd instead.")

            return self.validate_caps(self.dev.hd_caps)

    def validate_caps(self, caps):
            if caps:
                return caps
            else:
                print("Resolution not supported. Using "
                      "{}x{} instead.".format(self.dev.default_caps.width,
                                              self.dev.default_caps.height))
                return self.dev.default_caps

    def get_config(self):
        if self.video_fwk == "gstreamer":
            return self.gstreamer_config()
        elif self.video_fwk == "opencv":
            return self.opencv_config()
        else:
            if self.video_fwk != "v4l2":
                print("Invalid video framework. Using v4l2 instead.")

            return self.v4l2_config()

    def gstreamer_config(self):
        if self.video_src and os.path.isfile(self.video_src):
            sink_pipeline = set_appsink_video_pipeline(self.video_src)
            src_pipeline = set_appsrc_pipeline()
            return sink_pipeline, src_pipeline
        else:
            sink_pipeline = set_appsink_pipeline(device=self.dev.name)
            src_pipeline = set_appsrc_pipeline(width=self.dev_caps.width,
                                               height=self.dev_caps.height)
            return sink_pipeline, src_pipeline

    def opencv_config(self):
        if self.video_src and os.path.isfile(self.video_src):
            return cv2.VideoCapture(self.video_src), None
        else:
            dev = int(self.dev.name[10:])
            video = cv2.VideoCapture(dev)
            video.set(cv2.CAP_PROP_FRAME_WIDTH, self.dev_caps.width)
            video.set(cv2.CAP_PROP_FRAME_HEIGHT, self.dev_caps.height)

            return video, None

    def v4l2_config(self):
        if self.video_src and os.path.isfile(self.video_src):
            pipeline = v4l2_video_pipeline(self.video_src)
        else:
            pipeline = v4l2_camera_pipeline(width=self.dev_caps.width,
                                            height=self.dev_caps.height,
                                            device=self.dev.name,
                                            frate=self.dev_caps.framerate)

        return cv2.VideoCapture(pipeline), None


def save_image(image, dest, name):
    if not os.path.exists(dest):
        pathlib.Path(dest).mkdir(parents=True, exist_ok=True)

    count = 1
    img_name = os.path.join(dest, name+PNG)

    if os.path.exists(img_name):
        img_name = "{}({}){}".format(name, count, PNG)
        img_name = os.path.join(dest, img_name)

        while os.path.exists(img_name):
            count += 1
            img_name = "{}({}){}".format(name, count, PNG)
            img_name = os.path.join(dest, img_name)

    cv2.imwrite(img_name, image)
    print("Image saved as {}".format(img_name))
